// Program to send/receive UDP packets (and measure throughput)
// Bhaskaran Raman, bhaskar@cs.berkeley.edu, Nov 14 2000

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include <stdlib.h>
#include <assert.h>
#include <netinet/in.h>
#include <string.h>
#include <signal.h>

#include "Util.h"

#define PORT 2000
#define ALARM_TIME 5
/* The runtime of the program in ms */
#define RUN_TIME 120000

#define SEND_MODE 1
#define RECV_MODE 2
#define BOTH_MODE 3

struct hb_pkt {
  long seqno;
  char dummy[1424];
};
typedef struct hb_pkt hb_pkt_t;

/* The time at which heartbeats are started */
struct timeval start;
/* The socket descriptor */
int sock_fd;
/* Address on which to listen */
struct sockaddr_in listen_addr;
/* Port on which to listen */
int listen_port;
/* Address to which to send packets */
struct sockaddr_in to_addr;
/* Node to which to send packets */
char to_host[256];
/* Node to which to send packets */
struct in_addr to_ip;
/* Port to which to send packets */
int to_port;
/* The heart-beat interval (in ms) */
double hbInt;

long num_recd = 0;
long num_sent = 0;
double prev_stat_time = 0;
long num_lost = 0;
long num_dup = 0;
long next_seqno_expected = 0;
long num_rcv_err = 0;

/* Send heartbeat on the socket descriptor */
void sendHB(hb_pkt_t *pkt) {
  int len = sizeof(hb_pkt_t);
  int res = sendto(sock_fd, (void *)(pkt), len, 0,
		   (struct sockaddr *)&to_addr, sizeof(to_addr));
  if(res != len) {
    perror("sendto");
    return;
  }
} // End sendHB()

/* Try to receive heartbeat on the listen address */
bool tryReceive() {
  struct timeval t_o;
  t_o.tv_sec = t_o.tv_usec = 0; // don't wait around

  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(sock_fd, &rfds);
  select(sock_fd+1, &rfds, NULL, NULL, &t_o);

  /* Check if there is some data to be read */
  if(FD_ISSET(sock_fd, &rfds)) {
    int len = sizeof(hb_pkt_t);
    hb_pkt_t pkt;
    unsigned int length = sizeof(listen_addr);
    int res = recvfrom(sock_fd, (void *)(&pkt), len, 0,
		       (struct sockaddr*)&listen_addr, &length);
    if(0 && res != len) {
      //fprintf(stderr, "tryReceive recvfrom failed, res=%d\n", res);
      //perror("recvfrom");
      num_rcv_err++;
      return false;
    }
    num_recd++;
    int seqno_recd = ntohl(pkt.seqno);
    if(seqno_recd > next_seqno_expected) {
      num_lost += seqno_recd - next_seqno_expected;
      next_seqno_expected = seqno_recd+1;
    } else if(seqno_recd < next_seqno_expected) {
      num_dup += next_seqno_expected - seqno_recd;
    } else {
      next_seqno_expected = seqno_recd+1;
    }
    return true;
  } // End if(FD_ISSET)

  return false;
} // End tryReceive()

/* Print usage of program */
void usage(char *progname) {
  fprintf(stderr, "%s <-s/-r/-b> <port> <remote-host> <hbInt-in-ms> <bdcst=0/1>\n",
	  progname);
} // End usage()

void print_stats() {
  double now_time = now(start);
  double time_taken = now_time-prev_stat_time;
  double throughput = num_recd*8.0*sizeof(hb_pkt_t)/time_taken;
  throughput /= 1000.0; // In Mbps

  for(int i = 0; i < 60; i++) { fputc('-', stderr); }
  fprintf(stderr, "\n");
  fprintf(stderr, "Now time\t= %0.6f\n", now_time);
  fprintf(stderr, "Num pkts recd\t= %ld\n", num_recd);
  fprintf(stderr, "Num pkts sent\t= %ld\n", num_sent);
  fprintf(stderr, "Num pkts lost\t= %ld\n", num_lost);
  fprintf(stderr, "Num pkts dup\t= %ld\n", num_dup);
  fprintf(stderr, "Num pkts rcverr\t= %ld\n", num_rcv_err);
  fprintf(stderr, "Time taken\t= %0.4f ms\n", time_taken);
  fprintf(stderr, "Throughput\t= %0.6f Mbps\n", throughput);

  num_sent = num_recd = 0;
  num_lost = num_dup = 0;
  num_rcv_err = 0;
  prev_stat_time = now_time;
} // End print_stats()

void intr_catcher(int) {
  print_stats();
  fprintf(stderr, "Caught Ctrl-C, exiting...\n");
  exit(0);
} // End intr_catcher()

void alarm_catcher(int) {
  print_stats();
  if(now(start) > RUN_TIME) {
    fprintf(stderr, "Running over for %d ms. exiting...\n", RUN_TIME);
    exit(0);
  }
  alarm(ALARM_TIME);
} // End alarm_catcher()

int main(int argc, char *argv[]) {
	
  int isbroadcast = 0, rc, opt;

   /* Get command-line arguments */
  if(argc != 6) { usage(argv[0]); exit(11); }
  char mode_str[256];
  strcpy(mode_str, argv[1]);
  int port = atoi(argv[2]);
  strcpy(to_host, argv[3]);
  hbInt = atof(argv[4]);
  if(strcmp(argv[5], "bdcst=0") == 0) { isbroadcast = 0; }
  else if(strcmp(argv[5], "bdcst=1") == 0) { isbroadcast = 1; }
  else { usage(argv[0]); exit(12); }

  int mode = 0;
  if(strcmp(mode_str, "-s") == 0) { mode = SEND_MODE; }
  else if(strcmp(mode_str, "-r") == 0) { mode = RECV_MODE; }
  else if(strcmp(mode_str, "-b") == 0) { mode = BOTH_MODE; }
  else { usage(argv[0]); exit(2); }

  to_port = listen_port = port;

  /* socket */
  sock_fd = socket(PF_INET, SOCK_DGRAM, getprotobyname("udp")->p_proto);
  if(sock_fd < 0) { perror("socket failed"); exit(2); }
  fprintf(stderr, "socket successful\n");
  /* set socket option for broadcast */
  if(isbroadcast == 1){
    rc = setsockopt(sock_fd, SOL_SOCKET, SO_BROADCAST, (void*)&opt, sizeof(opt));
    if(rc<0) {
      perror("setsockopt");
      fprintf(stderr, "Can not set socket option\n");
      exit(4);
    }
    fprintf(stderr, "setsockopt successful\n");
  }
  /* bind */
  bzero((char *)&listen_addr, sizeof(listen_addr));
  listen_addr.sin_family =  AF_INET;
  listen_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  listen_addr.sin_port = htons(listen_port);

  if(bind(sock_fd, (struct sockaddr *)&listen_addr,
	  sizeof(listen_addr)) < 0) {
    fprintf(stderr, "Can't bind to port %d\n", listen_port);
    perror("bind failed"); exit(3);
  }
  fprintf(stderr, "bind successful\n");

  // Activate interrupt catcher
  signal(SIGINT, intr_catcher);
  signal(SIGTERM, intr_catcher);

  signal(SIGALRM, alarm_catcher);
  alarm(ALARM_TIME);

  /* init to_addr */
  struct hostent *server = gethostbyname(to_host);
  bzero((char *)&to_addr, sizeof(to_addr));
  bcopy((char *)server->h_addr, (char *)&to_addr.sin_addr.s_addr,
	server->h_length);
  to_addr.sin_family = PF_INET;
  to_addr.sin_port = htons(to_port);

  /* Forever, keep sending and receiving heartbeats */
  gettimeofday(&start, NULL);
  double startTime = 1.0*start.tv_sec + 1e-6 * start.tv_usec;
  num_recd = 0;
  prev_stat_time = 0;
  fprintf(stderr, "Start = %f\n", startTime);

  long seqno = 0;
  hb_pkt_t pkt;
  for(;;) {
    double now_time = now(start);
    // Check if time to send next heartbeat
    if(((mode == SEND_MODE) || (mode == BOTH_MODE)) &&
       (now_time >= (seqno*hbInt))) {
      for(int i = 0; i < 10; i++) {
	pkt.seqno = htonl(seqno);
	sendHB(&pkt);
	seqno++;
	num_sent++;
      }
    }
    // Try receiving
    if((mode == RECV_MODE) || (mode == BOTH_MODE)) {
      for(int i = 0; i < 10; i++) {
	if(!tryReceive()) { break; }
      }
    }
    //usleep(5000); // sleep for 5ms
  } // End for(;;)

} // End main()
