/* This implements WP2P_LL -- the LLC for WP2P */

#include "wp2p-ll.h"

UpDownBuf::UpDownBuf() {
  for(int i = 0; i < blim_; i++) {
    // All entries are free (without a valid packet) to begin with
    buf[i].free = true;
  }
} // End UpDownBuf::UpDownBuf()


/** To compare sequence numbers that are 16 bits long, we need to
    account for the wrap around.  We assume that we always compare
    sequence numbers within a "small window" -- small in comparison to
    the range of the sequence number space (2^16). */

bool UpDownBuf::less16(uint16_t x, uint16_t y) {
  if(x == y) { return false; }
  uint32_t x1 = x;
  uint32_t y1 = y;
  uint32_t diff1 = (x1 < y1) ? (y1 - x1) : (x1 - y1);
  uint32_t diff2 = 0x00010000 - diff1;
  if(diff1 < (MAX_PKTS_IN_PHASE*4 + (uint32_t) blim_)) {
    return (x1 < y1);
  } else if(diff2 < (MAX_PKTS_IN_PHASE*4 + (uint32_t) blim_)) {
    return (y1 < x1);
  } else {
    fprintf(stderr, "x1=%d, y1=%d, diff1=%d, diff2=%d, MAX_DIFF=%d\n",
	    (int) x1, (int) y1, (int) diff1, (int) diff2, (MAX_PKTS_IN_PHASE*4+blim_));
    exit(40);
  }
} // End less16()


bool UpDownBuf::addPkt(uint16_t dataseq, Packet *p, Handler *h) {
  // Get a free entry in 'buf', and add given packet there
  for(int i = 0; i < blim_; i++) {
    if(buf[i].free) {
      buf[i].seqno = dataseq;
      buf[i].p = p;
      buf[i].h = h;
      buf[i].free = false;
      buf[i].num_tx = 0;
      buf[i].num_phases_in_buf = 0;
      return true;
      
    }
  }
  return false; // No free entry available
} // End UpDownBuf::addPkt()


int UpDownBuf::minSeqNo(uint16_t *min_seq_p, int *num_phases_in_buf_p, int *num_tx_p) {

  int min_index = -1;
  uint16_t min_seq = 0;
  for(int i = 0; i < blim_; i++) {
    if(buf[i].free) { continue; }
    if((min_index == -1) || less16(buf[i].seqno, min_seq)) {
      min_index = i;
      min_seq = buf[i].seqno;
    }
  }

  if(min_index == -1) { return -1; }
  else {
    *min_seq_p = min_seq;
    if(num_phases_in_buf_p != NULL) {
      *num_phases_in_buf_p = buf[min_index].num_phases_in_buf;
    }
    if(num_tx_p != NULL) {
      *num_tx_p = buf[min_index].num_tx;
    }
    return min_index;
  }

} // End UpDownBuf::minSeqNo()


bool UpDownBuf::removeMin(Packet_p *pp, Handler_p *hp) {
  uint16_t min_seqno = 0;
  int min_index = minSeqNo(&min_seqno, NULL, NULL);
  if(min_index == -1) { return false; }
  *pp = buf[min_index].p;
  *hp = buf[min_index].h;
  buf[min_index].free = true;
  return true;
} // End UpDownBuf::removeMin()


int UpDownBuf::indexOf(uint16_t seqno) {
  for(int i = 0; i < blim_; i++) {
    if(!buf[i].free && (buf[i].seqno == seqno)) {
      return i;
    }
  }
  return -1;
} // End UpDownBuf::indexOf()


bool UpDownBuf::has_seqno(uint16_t seqno) {
  return (indexOf(seqno) != -1);
} // End UpDownBuf::has_seqno()


bool UpDownBuf::full() {
  for(int i = 0; i < blim_; i++) {
    if(buf[i].free) { return false; }
  }
  return true;
} // End UpDownBuf::full()


bool UpDownBuf::remove(uint16_t seq, Packet_p *pp, Handler_p *hp) {
  int index = indexOf(seq);
  if(index == -1) { return false; }
  else {
    *pp = buf[index].p;
    *hp = buf[index].h;
    buf[index].free = true;
    return true;
  }
} // End UpDownBuf::remove()


void UpDownBuf::sort_buf() {
  for(int i = 0; i < blim_; i++) {
    for(int j = i+1; j < blim_; j++) {

      // Note: this will end up sorting in "increasing" order of
      // sequence number, and all free 'buf' entries will be pushed
      // towards the end
      if(!buf[j].free &&
	 (buf[i].free || less16(buf[j].seqno, buf[i].seqno))) {
	pkt_info_t temp = buf[i];
	buf[i] = buf[j];
	buf[j] = temp;
      }

    } // End for(j)
  } // End for(i)
} // End UpDownBuf::sort_buf()


int UpDownBuf::getData(uint16_t num_pkts, Packet_p *P, Handler_p *H,
		       uint16_t *sendseq, int dummy_size) {

  // Sort so that we send packets in increasing order
  // Note: this strategy essentially means that if all ACKs are lost
  // in a PHASE, in the next phase, we will end up sending the same
  // set of packets as the earlier phase
  sort_buf();

  int num_done = 0;
  for(int i = 0; i < blim_; i++) {
    if(num_done == num_pkts) { break; }
    if(buf[i].free) { continue; }

    P[num_done] = buf[i].p;
    H[num_done] = buf[i].h;
    sendseq[num_done] = buf[i].seqno;
    buf[i].num_tx++;
    num_done++;
  }

  int num_valid = num_done;

  while(num_done < num_pkts) {
    //fprintf(stderr, "l-d %0.6f num_valid %d < num_pkts %d : gen fillers\n",
    //    Scheduler::instance().clock(),num_valid,num_pkts);
    // Generate a new packet (for filler)
    P[num_done] = Packet::alloc();
    // Note: Need to set the packet size here...
    // Mark the filler indication, so that the receiving end will know
    // that it is a filler/dummy packet
    HDR_CMN(P[num_done])->filler() = 1; // mark as filler packet
    HDR_CMN(P[num_done])->size() = dummy_size; // set size
    H[num_done] = NULL;
    sendseq[num_done] = 0; // seq. num. is irrelevant for filler pkts
    num_done++;
  }

  return num_valid;
} // End UpDownBuf::getData()


void UpDownBuf::incr_phase_count() {
  for(int i = 0; i < blim_; i++) {
    if(buf[i].free) { continue; }
    buf[i].num_phases_in_buf++;
  }
} // End UpDownBuf::incr_phase_count()


static class WP2P_LLClass : public TclClass {
public:
  WP2P_LLClass() : TclClass("LL/WP2P_LL") {}
  TclObject* create(int, const char*const*) {
    return (new WP2P_LL);
  }
} WP2P_class_ll;

/** The LLC Algorithm:

    The LLC algorithm description is as follows.  A PHASE consists of
    sending a bunch of packets and receiving a bunch of packets.  It
    is assumed at this stage that all packets are of the same length.
    When packets are not available to be sent, dummy/filler packets
    are generated and sent, which are dropped at the receiving end
    (after the relevant ack info in those dummy/filler packets are
    processed).

    MAX_RETRY is the limit on the number of retransmissions.

    Sequence number is a 16 bit field, with wrap around (to be taken
    care of while making comparisons between sequence numbers).

    In each round/phase, the "smallest" sequence numbers that have not
    yet been acked are sent.  Note that this may lead to a case where
    the same set of packets as the previous round are sent, if all
    acks are lost -- I implemented this since it was easier to
    implement.  Also I think the performance will be as good so long
    as the probability of all acks being lost is small.

    Ack has two parts, an 'ackseq' and an 'ackwin'.  The value of
    'ackseq' means that all sequence numbers (strictly) "less than"
    'ackseq' need no longer be retransmitted.  This need not
    necessarily mean that all these seq nos have been received.  It
    may happen that the other end waits for MAX_RETRY retransmissions
    of a 'hole', and decides to wait no longer and passes up
    everything beyond the 'hole' UP the stack.

    The 'ackwin' has 32 bits, and the i'th bit (i=0..31) is an ack bit
    for ackseq+i (addition with wrap around taken care of). */

WP2P_LL::WP2P_LL() : LL() {

  // Variable initializations
  recvseq = 0;
 
  seqno_ = 0;
  left_win_ = pred16(0);

  higher_layer_unblocked = true;

  dummyPktSize_ = 0;
  downQ.blim_ = MAX_PKTS;
  upQ.blim_ = MAX_PKTS_IN_PHASE*4;

} // End WP2P_LL::WP2P_LL()


int WP2P_LL::command(int argc, const char*const* argv) {
  if (argc == 3) {
    if (strcmp(argv[1], "set-my-node-id") == 0) {
      set_my_node_id(atoi(argv[2]));
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-nbr-node-id") == 0) {
      set_nbr_node_id(atoi(argv[2]));
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-my-mac-id") == 0) {
      my_mac_addr_ = atoi(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-nbr-mac-id") == 0) {
      nbr_mac_addr_ = atoi(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-dummy-pkt-size") == 0) {
      dummyPktSize_ = atoi(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-mac") == 0) {
      mac_ = (WP2P_MAC*) TclObject::lookup(argv[2]);
      return (TCL_OK);
    } else if (strcmp(argv[1], "set-limit") == 0) {
      downQ.blim_ = atoi(argv[2]);
      if(downQ.blim_ > MAX_PKTS){
	fprintf(stderr," The limit should be below %d \n",MAX_PKTS);
	exit(34);
      }
      return (TCL_OK);
    } 
  }
  return LL::command(argc, argv);
} // End WP2P_LL::command()


void WP2P_LL::recv(Packet* p, Handler* h) {
  /* Sanity Check */
  if(!initialized()) { fprintf(stderr, "ERR: not initialized\n"); exit(2); }

  hdr_cmn *ch = HDR_CMN(p);
  
  // If direction = UP, then pass it up the stack
  // Otherwise, set direction to DOWN and enqueue the packet
  if(ch->direction() == hdr_cmn::UP) {
    //fprintf(stderr, "l-d %0.6f %d:%d Pkt recv from mac\n",
    //    Scheduler::instance().clock(), my_node_id(), nbr_node_id());
    sendUp(p, h);
    return;
  }
  
  fprintf(stderr, "l-d %0.6f %d:%d %d Pkt recv from higher layers\n",
	  Scheduler::instance().clock(), my_node_id(), nbr_node_id(),ch->size());
  
  // Try to clear downQ buffer before adding packets to it.  Purge
  // pkts that have already been retransmitted too many times.  Here
  // we assume that the smaller sequence numbers have been transmitted
  // more often (which is true under the current LLC retransmission
  // mechanisms)
  for(;;) {
    uint16_t min_seqno;
    int num_tx;
    int min_index = downQ.minSeqNo(&min_seqno, NULL, &num_tx);
    if(min_index == -1) { break; } // no more pkts left
    if(num_tx < (MAX_RETRY+1)) {
      /* Still not reached MAX_RETRY count for the number of
	 retransmissions. */
      break;
    }
    Packet *p1;
    Handler *h1;
    downQ.removeMin(&p1, &h1); // remove the min seqno pkt
    fprintf(stderr, "%0.6f %d:%d %d D-DMR LL\n",
	    Scheduler::instance().clock(), my_node_id(), nbr_node_id(),min_seqno);
    drop(p1); // and drop it
  }

  ch->direction() = hdr_cmn::DOWN;
  bool add_success = downQ.addPkt(seqno_, p, h);
  if(!add_success) { fprintf(stderr, "%0.6f %d:%d - D-DBL LL\n",Scheduler::instance().clock(), my_node_id(), nbr_node_id()); drop(p); return; }
  fprintf(stderr, "%0.6f %d:%d %d D-ADD LL\n",
	  Scheduler::instance().clock(), my_node_id(), nbr_node_id(),seqno_);
  seqno_ = succ16(seqno_);

} // End WP2P_LL::recv()


void WP2P_LL::sendUp(Packet* p, Handler *h) {
  if (HDR_CMN(p)->error() > 0) {
    // Packet was in error, needs to be dropped and ignored
    fprintf(stderr, "%0.6f %d:%d %d U-DER LL\n",
	    Scheduler::instance().clock(), my_node_id(), nbr_node_id(), get_seqno(p));
    drop(p);
    return;
  }
  hdr_mac2P *llh = HDR_MAC2P(p);
  hdr_cmn *ch = HDR_CMN(p);

  /* Adjust the packet size - ie; strip off the header */
  ch->size() -= ETHER_HDR_LEN11;
  if(ch->size() < 0) {
    fprintf(stderr, "sendUp.err %0.6f %d:%d sz=%d seq=%d(%d) tcp_seq=%d\n",
	    Scheduler::instance().clock(), my_node_id(), nbr_node_id(),
	    ch->size(), get_seqno(p), ch->filler(), HDR_TCP(p)->seqno());
    //exit(89);
  }
  ch->num_forwards() += 1;

  // 2 bytes for seq (lseq) and 4 bytes for window (win)
  uint16_t *p1 = (uint16_t*)(llh->dh_ackwin);
  uint32_t *p2 = (uint32_t*)(llh->dh_ackwin+2);
  uint16_t ackseq = ntohs(*p1);
  uint32_t ackwin = ntohl(*p2);

  /* Keep track of the received ack information (so far as I see, this
     is not used anywhere) */
  R_info[recvseq].ackseq = ackseq;
  R_info[recvseq].ackwin = ackwin;
  recvseq++;

  // convert u_char to seqno;
  uint16_t dataseq = get_seqno(p);

  // Process ack info in recv pkt -- this will update the downQ
  processRecdAck(ackseq, ackwin);

  // process data info -- this will update the upQ
  processRecdData(dataseq, p, h);

} // End WP2P_LL::sendUp()


void WP2P_LL::processRecdAck(uint16_t ackseq, uint32_t ackwin){
  /* ackseq is the other end's left_win_+1, which means that the prior
     packets need not be retransmitted (correct?).  This is because
     when the other end has already passed UP a sequence number, say
     N, it means that either (a) N has been received, or (b) the other
     side has waited for MAX_RETRY retransmissions of N without
     success.  Hence it does not make sense to retransmit N (or
     sequence numbers "less than" N) any longer -- these can be
     deleted from downQ. */
  // Remove all packets strictly "less than" ackseq
  for(;;) {
    uint16_t min_seqno;
    int min_index = downQ.minSeqNo(&min_seqno, NULL, NULL);
    if(min_index == -1) {
      //fprintf(stderr,"no pkts \n");
      break;
    } // no more pkts left
    // No more pkts with seqno "less than" ackseq
    if(!downQ.less16(min_seqno, ackseq)) { break; }

    Packet *p1;
    Handler *h1;
    downQ.removeMin(&p1, &h1); // remove the min seqno pkt
    Packet::free(p1); // and free the allocation
    fprintf(stderr, "%0.6f %d:%d %d D-DAMR LL\n",
	    Scheduler::instance().clock(), my_node_id(), nbr_node_id(), min_seqno);
  }
  /** Remove all packets that have been acked from the downQ -- these
      don't need to be retransmitted anymore. */
  uint16_t seqno_acked = ackseq;
  uint32_t ackwin_shift = ackwin;

  for(int i = 0; i < 32; i++) {
    uint32_t ack_bit = (ackwin_shift & 0x00000001);
    if(ack_bit && downQ.has_seqno(seqno_acked)) {
      // Packet has been acked, can be removed from downQ
      Packet *p1;
      Handler *h1;
      bool remove_succ = downQ.remove(seqno_acked, &p1, &h1); // remove from buf
      if(!remove_succ) { fprintf(stderr, "downQ.remove failed\n"); exit(4); }
      Packet::free(p1); // and free the allocation
      fprintf(stderr, "%0.6f %d:%d %d D-DACK LL\n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(), seqno_acked);
    }
    seqno_acked = succ16(seqno_acked);
    ackwin_shift >>= 1;
  }
} // End WP2P_LL::processRecdAck()


void WP2P_LL::processRecdData(uint16_t dataseq, Packet *p, Handler *h){

  // Either add the given packet to the upQ, or, if it is a filler
  // packet, ignore it
  if(HDR_CMN(p)->filler() == 0) {
    // The packet should neither be in the upQ currently, nor should
    // it be something I have already passed up
    if(!upQ.has_seqno(dataseq) && upQ.less16(left_win_, dataseq)) {
      // Note: changing direction before sending it up to IP layer.
      // This way, if the IP layer sends the same pkt down via some
      // other interface's LL, that LL will see the pkt's direction as
      // DOWN.
      HDR_CMN(p)->direction() = hdr_cmn::DOWN;

      bool add_succ = upQ.addPkt(dataseq, p, h);
      if(!add_succ) { fprintf(stderr, "upQ.addPkt failed\n"); exit(5); }
      fprintf(stderr, "%0.6f %d:%d %d U-ADD LL\n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(), dataseq);
      /*      fprintf(stderr, "%0.6f %d:%d \n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(),
	      HDR_CMN(p)->size(), dataseq, HDR_TCP(p)->seqno());*/
    }
    else {
      fprintf(stderr, "%0.6f %d:%d %d U-DDUP LL\n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(), dataseq);
      Packet::free(p);
    }
  } else {
    Packet::free(p);
  }

  // Also, pass UP all possible pkts: those that have arrived in sequence
  for(;;) {
    uint16_t min_seqno;
    int num_phases_in_buf;
    int min_index = upQ.minSeqNo(&min_seqno, &num_phases_in_buf, NULL);
    if(min_index == -1) { break; } // no more pkts left
    uint16_t expected_seqno = succ16(left_win_);
    if(upQ.less16(expected_seqno, min_seqno) &&
       (num_phases_in_buf < (MAX_RETRY+1))) {
      /* Still waiting for a 'hole', and have been waiting in buf for
	 less than MAX_RETRY+1 phases (MAX_RETRY+1 is the max number
	 of tries, as opposed to retries). */
      break;
    }

    Packet *p1;
    Handler *h1;
    upQ.removeMin(&p1, &h1); // remove the min seqno pkt
    if(num_phases_in_buf == (MAX_RETRY+1)){
      fprintf(stderr, "%0.6f %d:%d %d U-PASS-MR LL\n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(),min_seqno);
    }
    else{
      fprintf(stderr, "%0.6f %d:%d %d U-PASS LL\n",
	      Scheduler::instance().clock(), my_node_id(), nbr_node_id(),min_seqno);
    }
    uptarget_->recv(p1, h1); // and pass it up
    if(!upQ.less16(left_win_, min_seqno)) { fprintf(stderr, "less16 failed\n"); exit(6); }
    left_win_ = min_seqno; // this would advance left_win_
  }
} // End WP2P_LL::processRecdData()


void WP2P_LL::processSendData() {
  
  // To be executed on Interrupt from mac asking for data

  // Decide on which pkts to send down -- get info from downQ
  Packet_p P[MAX_PKTS_IN_PHASE];
  Handler_p H[MAX_PKTS_IN_PHASE];
  uint16_t sendseq[MAX_PKTS_IN_PHASE];
  int num_valid = downQ.getData(mac_->send_n_, P, H, sendseq, dummyPktSize_);

  // Decide the ACK fields for sending -- get info from upQ
  uint16_t send_ackseq;
  uint32_t send_ackwin;
  processSendAck(&send_ackseq, &send_ackwin);

  // Debug messages
  /*fprintf(stderr, "%0.6f %d:%d LLC send_n_=%d num_valid=%d [", Scheduler::instance().clock(), my_id(), lnk_nbr_id(), mac_->send_n_, num_valid);
    for(int i = 0; i < mac_->send_n_; i++) {
    fprintf(stderr, "%d(%d)", sendseq[i], HDR_CMN(P[i])->size());
    if(i < (mac_->send_n_-1)) { fputc('|', stderr); }
    }*/
  //fprintf(stderr, "]\n");
  
  // Send the data to MAC
  for(int i = 0; i < mac_->send_n_; i++) {
    sendDown(sendseq[i], P[i], H[i], send_ackseq, send_ackwin);
  }

  recvseq = 0; // reset recvseq for next round of receiving

  /* Increment the phase count (by one) for all packets in the upQ */
  upQ.incr_phase_count();

} // End WP2P_LL::processSendData()
  

void WP2P_LL::processSendAck(uint16_t *send_ackseq, uint32_t *send_ackwin) {
  /* Simply set '*send_ackseq' to left_win_+1, meaning that sequence
     number 'left_win_' and its preceding sequence numbers have
     already been passed UP (after waiting for MAX_RETRY
     retransmissions), and hence these are as good as ACKed. */
  *send_ackseq = left_win_+1;

  /** Starting from '*send_ackseq', send info about which pkts have
      been received, and which have not been received so far.  Note:
      '*send_ackseq' itself would not have been received (since it is
      left_win_+1), hence the first bit is always '0' as implemented
      below. */
  uint16_t temp_ackseq = *send_ackseq;
  *send_ackwin = 0;
  for(int i = 0; i < 32; i++) {
    uint32_t ack_bit = upQ.has_seqno(temp_ackseq) ? 0x00000001 : 0x00000000;
    *send_ackwin |= (ack_bit << i);
    temp_ackseq = succ16(temp_ackseq);
  }
} // End WP2P_LL::processSendAck()


void WP2P_LL::sendDown(uint16_t sendseq, Packet *p, Handler *h,
		       uint16_t send_ackseq, uint32_t send_ackwin) {

  // If the packet is not dummy, make a copy of the packet -- need a
  // separate copy for each transmission.  The original packet will be
  // freed when the packet has been removed from downQ.
  Packet *p_copy = HDR_CMN(p)->filler() ? p : p->copy();


  // Fill the appropriate headers and send to downtarget_
  hdr_cmn* ch = HDR_CMN(p_copy);
  if(ch->size() <= dummyPktSize_)
    ch->size() = dummyPktSize_;
  else{
    fprintf(stderr," LL %f %d:%d Pkt size is greater than dummy, exiting \n",Scheduler::instance().clock(), my_node_id(), nbr_node_id());
    exit(890);
  }
  ch->size() += ETHER_HDR_LEN11; // Adjust the packet size
    
  /* Fill the MAC header fields */
  struct hdr_mac2P* dh = HDR_MAC2P(p_copy);

  dh->dh_fc.fc_protocol_version = MAC_ProtocolVersion;
  dh->dh_fc.fc_type       = MAC_Type_Data;
  dh->dh_fc.fc_subtype    = MAC_Subtype_Data;

  dh->dh_fc.fc_to_ds      = 0;
  dh->dh_fc.fc_from_ds    = 0;
  dh->dh_fc.fc_more_frag  = 0;
  dh->dh_fc.fc_retry      = 1; // should be 1 irrespective of number of retx
  dh->dh_fc.fc_pwr_mgt    = 0;
  dh->dh_fc.fc_more_data  = 0;
  dh->dh_fc.fc_wep        = 0;
  dh->dh_fc.fc_order      = 0;

  /* Send and receive addresses (not used so far as I can see) */
  //bcopy(lnk_nbr_mac_addr_, dh->dh_da, sizeof(lnk_nbr_mac_addr_));
  //bcopy(my_mac_addr_, dh->dh_sa, sizeof(my_mac_addr_));
  int dst = nbr_mac_addr_;
  int src = my_mac_addr_;
  if(dst > -2)
    STORE4BYTE(&dst, (dh->dh_da));
  if(src > -2)
    STORE4BYTE(&src, (dh->dh_sa));
  //fprintf(stderr,"LL dst %d %d \n", ETHER_ADDR(dh->dh_da),nbr_mac_addr_);
  /* The ack sequence number and the ack window */
  uint16_t *p1 = (uint16_t*)(dh->dh_ackwin);
  uint32_t *p2 = (uint32_t*)(dh->dh_ackwin+2);
  *p1 = htons(send_ackseq);
  *p2 = htonl(send_ackwin);

  /* The sequence number of the packet being sent */
  uint16_t *p3 = (uint16_t*)(dh->dh_bssid);
  *p3 = htons(sendseq);

  fprintf(stderr, "%0.6f %d:%d %d D-PASS%d LL\n",
	  Scheduler::instance().clock(), my_node_id(), nbr_node_id(), sendseq, ch->filler());
  downtarget_->recv(p_copy, h);
} // End WP2P_LL::sendDown()
