#ifndef WP2P_MAC_H
#define WP2P_MAC_H

#include <bi-connector.h>
#include <packet.h>
#include "mac-satcounter.h"
#include "marshall.h"
#include "phy.h"

/** A phase involves transmission in one direction, followed by
    transmission in the other direction.  MAX_PKTS_IN_PHASE defines
    the limit on the sum of the number of packets in either
    direction. */
#define MAX_PKTS_IN_PHASE 8

// The maximum number of interfaces (links/degree) a node can have
#define MAX_IFA 8

//captures current state of a link
enum RxTxState_e { IDLE='i', SYNRX='r', SYNTX='t' };
enum SyncState_e { ASYNC='a', SYNCING='g', INSYNC='s' };

typedef Packet *Packet_p;

#define DSSS_PreambleLength		144		// 144 bits
#define DSSS_PLCPHeaderLength		48		// 48 bits
#define DSSS_PLCPDataRate		1.0e6		// 1Mbps


#define ETHER_ADDR(x)   (GET4BYTE(x))
#define ETHER_ADDR_LEN  6
#define ETHER_TYPE_LEN  2
#define ETHER_FCS_LEN   4


#define MAC_ProtocolVersion     0x00
#define MAC_Type_Data           0x02
#define MAC_Subtype_Data        0x00

//#define TXRATE (11e6/8.0)
//#define PHASE_DUR_BYTES (send_n_ * ll_->get_pkt_size())
//#define PHASE_DUR (send_n_ * txtime(ll_->get_pkt_size(),bandwidth_))
#define TIMEOUT_VAL (1.25*phase_dur)

//R_MIN and R_MAX are the limits of processing delay. As of now we
// just take the average. In future, this can be random value between
// the limits
#define R_MIN 40e-6 // 40 micro-sec min delay
#define R_MAX 70e-6 // 70 micro-sec max delay
#define BUMP_DUR 100e-6 // 100 micro-sec bump duration
#define TX_WAIT_DUR 1e-6 //1 micro-sec

#define TIMEOUT_COUNT_MAX 20
#define ASYNC_COUNT_MAX  4
#define BUMP_COUNT_MAX  3
#define ROUNDS_BETN_BUMPS  50


struct frame_control {
        u_char          fc_subtype              : 4;
        u_char          fc_type                 : 2;
        u_char          fc_protocol_version     : 2;
                                                                                
        u_char          fc_order                : 1;
        u_char          fc_wep                  : 1;
        u_char          fc_more_data            : 1;
        u_char          fc_pwr_mgt              : 1;
        u_char          fc_retry                : 1;
        u_char          fc_more_frag            : 1;
        u_char          fc_from_ds              : 1;
        u_char          fc_to_ds                : 1;
};

struct hdr_mac2P {
        struct frame_control    dh_fc;
        u_int16_t               dh_duration;
        u_char                  dh_da[ETHER_ADDR_LEN];
        u_char                  dh_sa[ETHER_ADDR_LEN];
        u_char                  dh_bssid[ETHER_ADDR_LEN];
        u_char                  dh_ackwin[ETHER_ADDR_LEN];//CVR
        u_int16_t               dh_scontrol;
        u_char                  dh_body[0]; // XXX Non-ANSI
   // Header access methods
        static int offset_;
        inline static int& offset() { return offset_; }
        inline static hdr_mac2P* access(const Packet* p) {
                return (hdr_mac2P*) p->access(offset_);
        }
};

#define HDR_MAC2P(p) hdr_mac2P::access(p)

#define PLCP_HDR_LEN                            \
  	((DSSS_PreambleLength >> 3) +        \
 	 (DSSS_PLCPHeaderLength >> 3))

#define ETHER_HDR_LEN11				\
	(PLCP_HDR_LEN +				\
	 sizeof(struct hdr_mac2P) +		\
	 ETHER_FCS_LEN)


class WP2P_MAC;
class WP2P_LL;

/** This handler is used to schedule a future mac send event.  This is
    used during initialization (link-up), and also during switching from receive
    mode to send mode (after DSSS_RxTxTurnaroundTime) */
class WP2P_HandlerSend : public Handler {
public:
	WP2P_HandlerSend(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
 };

/** This handler is used to schedule packet send events to the PHY
    layer */
class WP2P_HandlerSend2Phy : public Handler {
public:
	WP2P_HandlerSend2Phy(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};

/** This handler is used to handle end of reception and to change from
    receive to transmit mode **/ 
class WP2P_HandlerEndRx : public Handler {
public:
  	WP2P_HandlerEndRx(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};

/** This handler is used to handle end of transmission **/
class WP2P_HandlerEndTx : public Handler {
public:
  	WP2P_HandlerEndTx(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};

/** This handler is used to handle timeout event **/
class WP2P_HandlerTimeOut : public Handler {
public:
  	WP2P_HandlerTimeOut(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};

/** This handler is ensures that we spend atleast enough time in RX mode **/
class WP2P_HandlerMinTimeInRx : public Handler {
public:
  	WP2P_HandlerMinTimeInRx(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};

/** This handler handles the timer for bumping **/
class WP2P_HandlerBumpTimeOut : public Handler {
public:
  	WP2P_HandlerBumpTimeOut(WP2P_MAC* m) : mac_(m) {}
	void handle(Event*);
protected:
	WP2P_MAC* mac_;
};


/** Defines the functionalities required for the point-to-point MAC */
class WP2P_MAC : public BiConnector {
  friend class WP2P_LL;
protected:
  int index_; //MAC addr
  /** Number of data pkts that will be sent in a PHASE */
  int send_n_;
  /** Number of data pkts that will be received in a PHASE */
  int recv_n_;
  /** currently received packet **/
  Packet *pktRx_;
  /** status of the packet **/
  int pktErr_;
  
  /** The duration involved in txing or recving required num of pkts **/
  double phase_dur;
  /** A pointer to the MAC of the neighbouring links at the same node */
  WP2P_MAC *ifa_nbr_mac_[MAX_IFA];
  int nbr_ifa;
  /** A pointer to the LL object at the same node */
  WP2P_LL *ll_;

  double bandwidth_; /* bandwidth of underlying link (bits/sec) */
  double delay_; /* line latency */
  
  Phy *netif_;            // network interface
  NsObject*       logtarget_;

  /** Num pkts received (in the UP direction) in the current PHASE */
  int num_rx;

  /** Num pkts recd from LL so far in this PHASE */
  int num_from_ll;
  /** To store packets from LL, before sending them to PHY.  This
      storage is needed because the interface between MAC/LL is such
      that a whole bunch of packets are gotten from LL in a single go.
      These are stored, and sent periodically */
  Packet_p P[MAX_PKTS_IN_PHASE];

  /** Packet transmission time 
      inline double txtime(Packet* p) {
      return 8.0 * ((HDR_CMN(p))->size()) / bandwidth_;
    }*/
  
  /** Is set if TO timer is active (to cancel if need be) */
  int timer;
  
  /** Is set is bump timer is active **/
  int bump_timer;
  /* Am I receiving on this link (note: I could be transmitting while
     also receiving -- collision) */
  bool receiving;
  /* If the header of the pkt is received during SYNTX mode, receiving
     is set to false as the node can't decode the TX. However to debug,
     dbg_recv is set to true as long as some pkt is being recv*/
  bool dbg_recv;
  /** Am I waiting to receive on this link?  This is set to false as
      soon as we start to receive on this link.  Note: we need both
      receiving and waiting_for_rx.  This is because waiting_for_rx
      could be set to false on first reception -- but there may be a
      temporary gap in reception, and we may still be receiving on the
      link, in which case we may need to wait before starting SYNTX. */
  bool waiting_for_rx;

  /** Number of times (rounds) I have timed out on this link.
      Incremented and checked each time I enter SYNTX.  If timed out
      more than TIMEOUT_COUNT_MAX, set link state to ASYNC. */
  SatCounter timeout_count;
  /** Number of times (rounds) I have been in ASYNC state.
      Incremented and checked each time I enter SYNTX.  If more than
      ASYNC_COUNT_MAX, bump self by BUMP_DUR. */
  SatCounter async_count;
  /** Number of times I have bumped myself trying to sync with one
      link.  If more than BUMP_COUNT_MAX, stop bumping self for this
      link.  Wait for ROUNDS_BETN_BUMPS rounds before trying to sync
      with this guy again, if necessary. */
  SatCounter bump_count;
  /** To count upto ROUNDS_BETN_BUMPS */
  SatCounter rounds_since_prev_bump;
  /** Should I try to bump on this link? */
  bool try_bump;

  enum RxTxState_e RxTxState; 
  enum SyncState_e SyncState;

  /** These are mainly node variables, so need to ensure that they are
      consistent across interfaces **/
  bool min_time_in_rx_done;
  
  /** Used to schedule a future MAC send event (after
      DSSS_RxTxTurnaroundTime) */
  WP2P_HandlerSend hSend;
  Event hSendEvent;

  /** Used to schedule individual packet send events to the PHY layer */
  WP2P_HandlerSend2Phy hSend2Phy;

  WP2P_HandlerEndRx hEndRx;

  WP2P_HandlerEndTx hEndTx;
  Event hEndTxEvent;

  WP2P_HandlerTimeOut hTimeOut;
  Event hTimeOutEvent;
  
  WP2P_HandlerMinTimeInRx hMinTimeInRx;
  Event hMinTimeInRxEvent;

  WP2P_HandlerBumpTimeOut hBumpTimeOut;
  Event hBumpTimeOutEvent;

  double txtime(Packet *p);
  double txtime(double psz, double drt);

public:
  /** Constructor, initialization */
  WP2P_MAC();
  /** TCL command interface */
  virtual int command(int argc, const char*const* argv);
  inline int addr() { return index_; }

  /** Send the given packet UP the stack */
  virtual void sendUp(Packet *p, Handler *h);
  /** Send the given packet DOWN the stack */
  virtual void sendDown(Packet *p, Handler *h);

  /** Called by the hSend handler */
  virtual void triggerSend();
  
  /** activate this link/interface **/
  virtual void link_up();

  /** deactivate this link/interface **/
  virtual void link_down();
  
  /** Before tx, update the state of the interface **/
  virtual void update_tx_state();
  /** Transmit on all the interfaces that are not idle/down **/
  virtual void tx_on_all_links(); 
  /** called by tx_on_all_links to handle individual links **/
  virtual void tx_on_link();
  /** called at end of reception of a packet **/
  virtual void end_rx();
  /** checks if all interfaces are done receiving, returns true if
      all are done receiving, else false **/
  virtual bool all_done_receiving();
  /** called at end of transmission of a packet **/
  virtual bool end_tx();
  /** called when a timeout happens on a interface due to not receving
      a packet **/
  virtual bool timeout();

  /** When all are done recv, this routine is called to see if 
      bumping is needed. If needed, bump timer is scheduled. If not,
      pkts are tx on all links**/
  virtual void checkBump();
  virtual bool ToBump();

  /** used to schedule a bump event. At timeout, tx on all links
      begins. **/
  virtual void bump_timeout();
  
  /** Bump timeout is scheduled on all links when a bump event is
      due. The timer is cancelled when a pkt is recv during the wait
      during bump_dur **/
  virtual double scheduleBumpTimer(double);
  virtual void cancelBumpTimer();

  /** We need to spend some minimum time in RX mode. This routine is
      called when that time is spent. We are now ready to change to TX
      mode if all interfaces are done receiving. **/
  virtual bool done_min_time_in_rx();
  virtual void cancelMTRDTimer();
  
  /** This procedure sends all packets stored in the array 'P' above,
      in the downward direction */
  virtual void sendAll();
  /** Send one packet down to PHY, called by hSend2Phy */
  virtual void send2Phy(Packet *p);

}; // End class WP2P_MAC

#endif
